@@ -1,4 +1,4 @@
-(* $Id: frontend.ml 239 2014-10-14 09:39:42Z gerd $
+(* $Id$
  * ----------------------------------------------------------------------
  *
  *)
@@ -10,7 +10,7 @@ exception Usage;;
 type mode =
     M_use | M_query | M_install | M_remove | M_compiler of string | M_dep
   | M_printconf | M_list | M_browser | M_call of (string*string)
-  | M_doc
+  | M_doc | M_lint
 ;;
 
 
@@ -28,19 +28,22 @@ and modifier =
 
 let slashify s =
   match Findlib_config.system with
-    | "mingw" | "cygwin" ->
-	let u = String.copy s in
-	for k = 0 to String.length u - 1 do
-	  if u.[k] = '\\' then u.[k] <- '/'
-	done;
-	u
+    | "mingw" | "mingw64" | "cygwin" ->
+        let b = Buffer.create 80 in
+        String.iter
+          (function
+            | '\\' -> Buffer.add_char b '/'
+            | c -> Buffer.add_char b c
+          )
+          s;
+        Buffer.contents b
     | _ ->
 	s
 
 
 let out_path ?(prefix="") s =
   match Findlib_config.system with
-    | "mingw" | "cygwin" ->
+    | "mingw" | "mingw64" | "cygwin" ->
 	let u = slashify s in
 	prefix ^ 
 	  (if String.contains u ' ' then
@@ -358,74 +361,76 @@ let run_command ?filter verbose cmd args =
   );
   flush stdout;
 
-  let filter_input, cmd_output =
-    match filter with
-	None -> Unix.stdin (* dummy *), Unix.stdout
-      | Some f -> Unix.pipe()
-  in
+  if verbose <> Only_show then (
+    let filter_input, cmd_output =
+      match filter with
+          None -> Unix.stdin (* dummy *), Unix.stdout
+        | Some f -> Unix.pipe()
+    in
 
-  (* Signals: On SIGINT, we wait until the subprocess finishes, and
-   * die then. This allows us to call interactive commands as subprocesses.
-   *)
+    (* Signals: On SIGINT, we wait until the subprocess finishes, and
+     * die then. This allows us to call interactive commands as subprocesses.
+     *)
 
-  let old_sigint =
-    Sys.signal Sys.sigint Sys.Signal_ignore in
+    let old_sigint =
+      Sys.signal Sys.sigint Sys.Signal_ignore in
 
-  let need_exe =
-    List.mem Findlib_config.system [ "win32"; "win64"; "mingw" ] in
+    let need_exe =
+      List.mem Findlib_config.system [ "win32"; "win64"; "mingw"; "mingw64" ] in
 
-  let fixed_cmd =
-    if need_exe then (
-      if Filename.check_suffix cmd ".exe" then cmd else cmd ^ ".exe" 
-    )
-    else
-      cmd in
+    let fixed_cmd =
+      if need_exe then (
+        if Filename.check_suffix cmd ".exe" then cmd else cmd ^ ".exe" 
+      )
+      else
+        cmd in
 
-  let pid =
-    Unix.create_process
-      fixed_cmd
-      (Array.of_list (cmd :: args))
-      Unix.stdin
-      cmd_output
-      Unix.stderr
-  in
+    let pid =
+      Unix.create_process
+        fixed_cmd
+        (Array.of_list (cmd :: args))
+        Unix.stdin
+        cmd_output
+        Unix.stderr
+    in
 
-  begin match filter with
-      Some filter_fun ->
-	begin
-	  Unix.close cmd_output;
-	  let ch = Unix.in_channel_of_descr filter_input in
-	  try
-	    while true do
-	      let line = input_line ch in
-	      match filter_fun line with
-		  None -> ()       (* Suppress line *)
-		| Some line' -> print_endline line'
-	    done;
-	    assert false
-	  with
-	      End_of_file ->
-		close_in ch;
-		flush stdout
-	end
-    | None -> ()
-  end;
+    begin match filter with
+        Some filter_fun ->
+          begin
+            Unix.close cmd_output;
+            let ch = Unix.in_channel_of_descr filter_input in
+            try
+              while true do
+                let line = input_line ch in
+                match filter_fun line with
+                    None -> ()       (* Suppress line *)
+                  | Some line' -> print_endline line'
+              done;
+              assert false
+            with
+                End_of_file ->
+                  close_in ch;
+                  flush stdout
+          end
+      | None -> ()
+    end;
 
-  let (_,status) = Unix.waitpid [] pid in
-  Sys.set_signal Sys.sigint old_sigint;
-  begin
-    match status with
-      Unix.WEXITED 0 -> ()
-    | Unix.WEXITED n ->
-	if verbose = Verbose then
-	  print_string (cmd ^ " returned with exit code " ^ string_of_int n ^ "\n");
-	exit n
-    | Unix.WSIGNALED _ ->
-	print_string (cmd ^ " got signal and exited\n");
-	exit 2
-    | Unix.WSTOPPED _ ->
-	failwith "Your operating system does not work correctly"
-  end
+    let (_,status) = Unix.waitpid [] pid in
+    Sys.set_signal Sys.sigint old_sigint;
+    begin
+      match status with
+        Unix.WEXITED 0 -> ()
+      | Unix.WEXITED n ->
+          if verbose = Verbose then
+            print_string (cmd ^ " returned with exit code " ^ string_of_int n ^ "\n");
+          exit n
+      | Unix.WSIGNALED _ ->
+          print_string (cmd ^ " got signal and exited\n");
+          exit 2
+      | Unix.WSTOPPED _ ->
+          failwith "Your operating system does not work correctly"
+    end
+  )
 ;;
 
 
@@ -668,6 +673,7 @@ let expand predicates eff_packages format =
     (* format:
      * %p         package name
      * %d         package directory
+     * %m         META file
      * %D         description
      * %v         version
      * %a         archive file(s)
@@ -684,6 +690,7 @@ let expand predicates eff_packages format =
 	 let spec =
 	   [ "%p",  [pkg];
              "%d",  [out_path dir];
+             "%m",  [out_path (package_meta_file pkg)];
 	     "%D",  [try package_property predicates pkg "description"
 		     with Not_found -> "[n/a]"];
 	     "%v",  [try package_property predicates pkg "version"
@@ -716,6 +723,7 @@ let help_format() =
 
     %p         package name
     %d         package directory
+    %m         META file
     %D         description
     %v         version
     %a         archive file(s)
@@ -1148,8 +1156,12 @@ let ocamlc which () =
   let threads_dir = Filename.concat stdlibdir "threads" in
   let vmthreads_dir = Filename.concat stdlibdir "vmthreads" in
 
-  let initl_file_needed =
+  let create_toploop =
     List.mem "create_toploop" !predicates && List.mem "findlib" eff_link in
+  let have_dynload =
+    List.mem "findlib.dynload" eff_link in
+  let initl_file_needed =
+    create_toploop || have_dynload in
 
   let initl_file_name =
     if initl_file_needed then
@@ -1169,21 +1181,22 @@ let ocamlc which () =
 		  0o777
 		  initl_file_name in
     try
+      List.iter
+        (fun pkg ->
+           Printf.fprintf
+             initl
+             "Findlib.record_package Findlib.Record_core %S;;\n"
+             pkg
+        )
+        eff_packages;
       output_string initl
-	("Topfind.don't_load [" ^
-	 String.concat ";"
-	   (List.map
-	      (fun pkg -> "\"" ^ String.escaped pkg ^ "\"")
-	      eff_link) ^
-	 "];;\n");
-      output_string initl
-	("Topfind.predicates := [" ^
+	("Findlib.record_package_predicates [" ^
 	 String.concat ";"
 	   (List.map
 	      (fun pred -> "\"" ^ String.escaped pred ^ "\"")
-	      ("toploop" :: 
-		 (List.filter (fun p -> p <> "create_toploop") !predicates))) ^
-	 "];;\n");
+              !predicates
+           ) ^
+	   "];;\n");
       close_out initl;
     with
       any ->
@@ -1192,7 +1205,7 @@ let ocamlc which () =
 	raise any
   end;
 
-  if initl_file_needed then
+  if initl_file_needed && verbose <> Only_show then
     at_exit
       (fun () ->
 	let tr f x = try f x with _ -> () in
@@ -1253,6 +1266,11 @@ let ocamlc which () =
 	 (fun pkg ->
 	   let al = try package_property !predicates pkg "archive"
 	            with Not_found -> "" in
+           let al_ext =
+             if have_dynload && pkg = "findlib.dynload" then
+               [ initl_file_name ]
+             else
+               [] in
 	   let pkg_dir =
 	     if pkg = "threads" then   (* MAGIC *)
 	       match !threads with
@@ -1265,11 +1283,11 @@ let ocamlc which () =
 	   List.map
 	     (fun arch ->
 		resolve_path ~base:pkg_dir arch)
-	     (Fl_split.in_words al)
+	     (Fl_split.in_words al @ al_ext)
 	 )
 	 eff_link)
     @
-    (if initl_file_needed then
+    (if create_toploop then
        [ initl_file_name ]
      else
        []
@@ -1766,7 +1784,7 @@ let copy_file ?(rename = (fun name -> name)) ?(append = "") src dstdir =
 		   outpath in
     try
       let buflen = 4096 in
-      let buf = String.create buflen in
+      let buf = String.create buflen in   (* FIXME: Bytes.create *)
       let pos = ref 0 in
       let len = ref (input ch_in buf 0 buflen) in
       while !len > 0 do
@@ -1931,6 +1949,19 @@ exception Skip_file;;
 
 type which = Auto | Dll | No_dll;;
 
+let meta_pkg meta_name =
+  let f = open_in meta_name in
+  try
+    let pkg = Fl_metascanner.parse f in
+    close_in f;
+    pkg
+  with
+  | Failure s
+  | Stream.Error s ->
+    close_in f;
+    failwith ("Cannot parse '" ^ meta_name ^ "': " ^ s)
+
+
 let install_package () =
   let destdir = ref (default_location()) in
   let metadir = ref (meta_directory()) in
@@ -2038,18 +2069,7 @@ let install_package () =
 	  else
 	    failwith "The META file is missing" in
 
-  let meta_pkg =
-    let f = open_in meta_name in
-    try
-      let pkg = Fl_metascanner.parse f in
-      close_in f;
-      pkg
-    with
-      | Failure s
-      | Stream.Error s ->
-	  close_in f;
-	  failwith ("Cannot parse '" ^ meta_name ^ "': " ^ s)
-  in
+  let meta_pkg = meta_pkg meta_name in
 
   if not !add_files then (
     (* Check for frequent reasons why installation can go wrong *)
@@ -2322,14 +2342,14 @@ let print_configuration() =
   in
 
   let var = ref None in
-  let errmsg = "usage: ocamlfind printconf (conf|path|destdir|metadir|stdlib|ldconf)" in
+  let errmsg = "usage: ocamlfind printconf (conf|path|destdir|metadir|effmetadir|stdlib|ldconf)" in
 
   parse_args
         []
 	(fun s ->
 	   if !var <> None then raise(Arg.Bad "Unexpected argument");
 	   match s with
-	       ("conf" | "path" | "destdir" | "metadir" | "stdlib" | "ldconf") ->
+	       ("conf" | "path" | "destdir" | "metadir" | "metapath" | "stdlib" | "ldconf") ->
 		 var := Some s
 	     | _ ->
 		 raise(Arg.Bad "Bad argument");
@@ -2364,6 +2384,11 @@ let print_configuration() =
 	print_endline (Findlib.default_location())
     | Some "metadir" ->
 	print_endline (Findlib.meta_directory())
+    | Some "metapath" ->
+        let mdir = Findlib.meta_directory() in
+        let ddir = Findlib.default_location() in
+	print_endline 
+          (if mdir <> "" then mdir ^ "/META.%s" else ddir ^ "/%s/META")
     | Some "stdlib" ->
 	print_endline (Findlib.ocaml_stdlib())
     | Some "ldconf" ->
@@ -2390,6 +2415,29 @@ let ocamlcall pkg cmd =
   run_command Normal path args
 ;;
 
+(** lint META file *)
+let lint () =
+
+  let meta_files = Queue.create () in
+
+  parse_args
+    ~align:false
+    ( Arg.align [
+        ])
+    (fun s -> if Sys.file_exists s
+      then Queue.add s meta_files
+      else raise(Arg.Bad (Printf.sprintf "%s: file doesn't exists" s)))
+    "usage: ocamlfind ocamldoc <options> <files>...";
+
+  let error =
+    Queue.fold (fun error file ->
+        let pkg = meta_pkg file in
+        let error = Fl_lint.warn pkg || error in
+        error
+      ) false meta_files in
+  exit (if error then 1 else 0)
+;;
+
 
 let rec select_mode () =
   let k = !Arg.current in
@@ -2411,6 +2459,7 @@ let rec select_mode () =
     | ("ocamldoc"|"-ocamldoc"|"doc")       -> incr Arg.current; M_doc
     | ("printconf"|"-printconf")           -> incr Arg.current; M_printconf
     | ("list"|"-list")                     -> incr Arg.current; M_list
+    | ("lint"|"-lint")                     -> incr Arg.current; M_lint
     | "-toolchain" ->
 	let t = try arg (k+2) with Not_found -> raise Usage in
 	Findlib.init ~toolchain:t ();
@@ -2454,6 +2503,7 @@ let main() =
     | M_doc            -> ocamldoc()
     | M_call(pkg,cmd)  -> ocamlcall pkg cmd
     | M_compiler which -> ocamlc which ()
+    | M_lint           -> lint()
   with
     Usage ->
       prerr_endline "Usage: ocamlfind query        [-help | other options] <package_name> ...";
@@ -2471,6 +2521,7 @@ let main() =
       prerr_endline "   or: ocamlfind install      [-help | other options] <package_name> <file> ...";
       prerr_endline "   or: ocamlfind remove       [-help | other options] <package_name>";
       prerr_endline "   or: ocamlfind printconf    [-help] [variable]";
+      prerr_endline "   or: ocamlfind lint         [-help] <file>";
       prerr_endline "   or: ocamlfind list";
       prerr_endline "   or: ocamlfind pkg/cmd arg ...";
       prerr_endline "Select toolchain with:";
